#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include <math.h>

#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64

#define OLED_RESET    -1
#define SCREEN_ADDRESS 0x3C

Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

//-----------------------------MAP----------------------------------------------
#define mapX  8      //map width
#define mapY  8      //map height
#define mapS 8      //map cube size
//#define PI 3.1415926535

#define BUTTON_PIN_FORWARD 7
#define BUTTON_PIN_BACKWARD 9
#define BUTTON_PIN_RIGHT 8
#define BUTTON_PIN_LEFT 10 


float px,py,pdx,pdy,pa;

int mapArray[]=           //the map array. Edit to change level but keep the outer walls
{
 1,1,1,1,1,1,1,1,
 1,0,1,0,0,0,0,1,
 1,0,1,0,0,0,0,1,
 1,0,1,0,0,0,0,1,
 1,0,0,0,0,0,0,1,
 1,0,0,0,0,1,0,1,
 1,0,0,0,0,0,0,1,
 1,1,1,1,1,1,1,1,  
};


void setup() {
  
  if (!display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS)) {
    Serial.println(F("SSD1306 allocation failed"));
    for (;;);
  }
  
  Serial.begin(9600);

  pinMode(BUTTON_PIN_RIGHT, INPUT_PULLUP); // Ustawienie pinu przycisku w prawo jako wejście z pull-up
  pinMode(BUTTON_PIN_LEFT, INPUT_PULLUP);
  pinMode(BUTTON_PIN_FORWARD, INPUT_PULLUP); // Ustawienie pinu przycisku w prawo jako wejście z pull-up
  pinMode(BUTTON_PIN_BACKWARD, INPUT_PULLUP);


//ustawienie początkowe
 px=30; py=40; pa=90;
 pdx=cos(degToRad(pa)); pdy=-sin(degToRad(pa)); 
 display.drawRect(px,py, 3, 3, WHITE);

  display.display(); // Wyświetl pustą tablicę pikseli (ekran początkowy)
  delay(2000);  // Poczekaj 2 sekundy
  display.clearDisplay();

}

void loop() {
  display.setRotation(2); 
  display.clearDisplay();
/*
//mapa 2d


  for (int y = 0; y < mapY; y++) {
    for (int x = 0; x < mapX; x++) {
      int xo, yo, if1;
      
      xo = x * mapS;
      yo = y * mapS;

      if1=y * mapX + x;
 
      if (mapArray[if1] == 1) {
        // Jeśli na mapie znajduje się przeszkoda (wartość 1), wyświetl piksel
        for (int i = 0; i < mapS; i++) {
          for (int j = 0; j < mapS; j++) {
            display.drawPixel(xo + i, yo + j, WHITE);
          }
        }
      } else {
        // W przeciwnym razie, wyświetl piksel w kolorze czarnym
        for (int i = 0; i < mapS; i++) {
          for (int j = 0; j < mapS; j++) {
            display.drawPixel(xo + i, yo + j, BLACK);
          }
        }
      }
    }
  }
  */

//ruszanie

if(digitalRead(BUTTON_PIN_LEFT) == LOW){ pa+=20; pa=FixAng(pa); pdx=cos(degToRad(pa)); pdy=-sin(degToRad(pa));}  
if(digitalRead(BUTTON_PIN_RIGHT) == LOW){ pa-=20; pa=FixAng(pa); pdx=cos(degToRad(pa)); pdy=-sin(degToRad(pa));} 

 int xo=0; if(pdx<0){ xo=-2;} else{ xo=2;}                                    //x offset to check map
 int yo=0; if(pdy<0){ yo=-2;} else{ yo=2;}                                    //y offset to check map
 int ipx=px/8.0, ipx_add_xo=(px+xo)/8.0, ipx_sub_xo=(px-xo)/8.0;             //x position and offset
 int ipy=py/8.0, ipy_add_yo=(py+yo)/8.0, ipy_sub_yo=(py-yo)/8.0;             //y position and offset
 
if(digitalRead(BUTTON_PIN_FORWARD) == LOW){
  if(mapArray[ipy*mapX        + ipx_add_xo]==0){ px+=pdx*5;}
  if(mapArray[ipy_add_yo*mapX + ipx       ]==0){ py+=pdy*5;}
}

if(digitalRead(BUTTON_PIN_BACKWARD) == LOW){
  if(mapArray[ipy*mapX        + ipx_sub_xo]==0){ px-=pdx*5;}
  if(mapArray[ipy_sub_yo*mapX + ipx       ]==0){ py-=pdy*5;}
}
/*
//wyswietlanie postaci 2d
 
 display.drawRect(px,py, 3, 3, WHITE);

  // Rysowanie linii reprezentującej kierunek gracza
  int endX = px + pdx *10;
  int endY = py + pdy * 10;
  display.drawLine(px, py, endX, endY, WHITE);
  */
  drawRays();

 /* char strForOLED[10];
  dtostrf(pa, 6, 2, strForOLED);
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE) ; // Draw white text
  display.setCursor(70, 15);
  display.cp437(true) ;
  display.write(strForOLED);*/
  
  display.display();
  delay(100);
}


float degToRad(int a) { return a*M_PI/180.0;}
int FixAng(int a){ if(a>359){ a-=360;} if(a<0){ a+=360;} return a;}
//float distance(ax,ay,bx,by,ang){ return cos(ang*PI/180.0)*(bx-ax)-sin(ang*PI/180.0)*(by-ay);} 

void drawRays(){
  
int r,mx,my,mp,dof,side; 
float vx,vy,rx,ry,ra,xo,yo,disV,disH; 

ra=FixAng(pa+60.0);
  for(r=0;r<120;r++){
//---Vertical--- 
  dof=0; side=0; disV=100000;
  float Tan=tan(degToRad(ra));
       if(cos(degToRad(ra))> 0.001){ rx=(((int)px>>3)<<3)+mapS;      ry=(px-rx)*Tan+py; xo= mapS; yo=-xo*Tan;}//looking left
  else if(cos(degToRad(ra))<-0.001){ rx=(((int)px>>3)<<3) -0.0001; ry=(px-rx)*Tan+py; xo=-mapS; yo=-xo*Tan;}//looking right
  else { rx=px; ry=py; dof=mapS;}                                                  //looking up or down. no hit  

  while(dof<8) 
  { 
   mx=(int)(rx)>>3; my=(int)(ry)>>3; mp=my*mapX+mx;                     
   if(mp>0 && mp<mapX*mapY && mapArray[mp]==1){ dof=mapS; disV=cos(degToRad(ra))*(rx-px)-sin(degToRad(ra))*(ry-py);}//hit         
   else{ rx+=xo; ry+=yo; dof+=1;}                                               //check next horizontal
  } 


  vx=rx; vy=ry;

  //---Horizontal---
  dof=0; disH=100000;
  Tan=1.0/Tan; 
       if(sin(degToRad(ra))> 0.001){ ry=(((int)py>>3)<<3) -0.0001; rx=(py-ry)*Tan+px; yo=-mapS; xo=-yo*Tan;}//looking up 
  else if(sin(degToRad(ra))<-0.001){ ry=(((int)py>>3)<<3)+mapS;      rx=(py-ry)*Tan+px; yo= mapS; xo=-yo*Tan;}//looking down
  else{ rx=px; ry=py; dof=mapS;}                                                   //looking straight left or right
 
  while(dof<8) 
  { 
   mx=(int)(rx)>>3; my=(int)(ry)>>3; mp=my*mapX+mx;                          
   if(mp>0 && mp<mapX*mapY && mapArray[mp]==1){ dof=mapS; disH=cos(degToRad(ra))*(rx-px)-sin(degToRad(ra))*(ry-py);}//hit         
   else{ rx+=xo; ry+=yo; dof+=1;}                                               //check next horizontal
  } 

  if(disV<disH){ rx=vx; ry=vy; disH=disV;}
 

  int ca=FixAng(pa-ra); disH=disH*cos(degToRad(ca)); 
  int lineH = (mapS*64)/(disH); if(lineH>64){ lineH=64;}                     //line height and limit
  int lineOff = 32 - (lineH>>1); 

  
  display.drawLine(r,lineOff,r,lineOff+lineH,WHITE);
  //display.drawLine(px,py,rx,ry,WHITE);

   /*   char strForOLED2[10];
  dtostrf(py, 6, 2, strForOLED2);
  display.setTextSize(1);
  display.setTextColor(SSD1306_WHITE) ; // Draw white text
  display.setCursor(70, 50);
  display.cp437(true) ;
  display.write(strForOLED2);*/

  ra=FixAng(ra-1); 
  }

}
